import type { Module, Submodule } from "../../module.ts";
import type { BaseAttributes, constraint, Default, Narrowed, of, Optional, Predicate } from "../inference.ts";
import { epoch } from "./epoch.ts";
import { integer } from "./integer.ts";
export declare const number: number.module;
export type AtLeast<rule> = {
    atLeast: constraint<rule>;
};
export type AtMost<rule> = {
    atMost: constraint<rule>;
};
export type MoreThan<rule> = {
    moreThan: constraint<rule>;
};
export type LessThan<rule> = {
    lessThan: constraint<rule>;
};
export type DivisibleBy<rule> = {
    divisibleBy: constraint<rule>;
};
export declare namespace number {
    type atLeast<rule> = of<number, AtLeast<rule>>;
    type moreThan<rule> = of<number, MoreThan<rule>>;
    type atMost<rule> = of<number, AtMost<rule>>;
    type lessThan<rule> = of<number, LessThan<rule>>;
    type divisibleBy<rule> = of<number, DivisibleBy<rule>>;
    type narrowed = of<number, Narrowed>;
    type optional = of<number, Optional>;
    type defaultsTo<rule> = of<number, Default<rule>>;
    type branded<rule> = of<number, Predicate<rule>>;
    interface ownConstraints extends AtLeast<number>, MoreThan<number>, LessThan<number>, AtMost<number>, DivisibleBy<number> {
    }
    interface Attributes extends BaseAttributes, Partial<ownConstraints> {
    }
    type NaN = branded<"NaN">;
    type Infinity = branded<"Infinity">;
    type NegativeInfinity = branded<"NegativeInfinity">;
    type safe = branded<"safe">;
    type is<attributes> = of<number, attributes>;
    type minSchemaToConstraint<schema, rule> = schema extends {
        exclusive: true;
    } ? MoreThan<rule> : AtLeast<rule>;
    type maxSchemaToConstraint<schema, rule> = schema extends {
        exclusive: true;
    } ? LessThan<rule> : AtMost<rule>;
    type applyAttribute<attribute> = attribute extends MoreThan<infer rule> ? moreThan<rule> : attribute extends AtLeast<infer rule> ? atLeast<rule> : attribute extends AtMost<infer rule> ? atMost<rule> : attribute extends LessThan<infer rule> ? lessThan<rule> : attribute extends DivisibleBy<infer rule> ? divisibleBy<rule> : attribute extends Optional ? optional : attribute extends Default<infer rule> ? defaultsTo<rule> : attribute extends Predicate<infer rule> ? branded<rule> : never;
    type module = Module<submodule>;
    type submodule = Submodule<$>;
    type $ = {
        root: number;
        epoch: epoch;
        integer: integer;
        safe: safe;
        NaN: NaN;
        Infinity: Infinity;
        NegativeInfinity: NegativeInfinity;
    };
}
