import type { Module, Submodule } from "../../module.ts";
import type { AtLeastLength, AtMostLength, Default, ExactlyLength, LessThanLength, MoreThanLength, Narrowed, Optional, Predicate, constraint, of } from "../inference.ts";
import { alpha } from "./alpha.ts";
import { alphanumeric } from "./alphanumeric.ts";
import { base64 } from "./base64.ts";
import { capitalize } from "./capitalize.ts";
import { creditCard } from "./creditCard.ts";
import { stringDate } from "./date.ts";
import { digits } from "./digits.ts";
import { email } from "./email.ts";
import { type stringInteger } from "./integer.ts";
import { ip } from "./ip.ts";
import { type stringJson } from "./json.ts";
import { lower } from "./lower.ts";
import { normalize } from "./normalize.ts";
import { type stringNumeric } from "./numeric.ts";
import { semver } from "./semver.ts";
import { trim } from "./trim.ts";
import { upper } from "./upper.ts";
import { url } from "./url.ts";
import { uuid } from "./uuid.ts";
export declare const string: Module<{
    root: unknown;
    trim: Submodule<trim.submodule>;
    normalize: Submodule<{
        root: unknown;
        NFC: Submodule<{
            root: unknown;
            preformatted: unknown;
        }>;
        NFD: Submodule<{
            root: unknown;
            preformatted: unknown;
        }>;
        NFKC: Submodule<{
            root: unknown;
            preformatted: unknown;
        }>;
        NFKD: Submodule<{
            root: unknown;
            preformatted: unknown;
        }>;
    }>;
    alpha: unknown;
    alphanumeric: unknown;
    url: Submodule<url.submodule>;
    json: Submodule<stringJson.submodule>;
    base64: Submodule<{
        root: unknown;
        url: unknown;
    }>;
    creditCard: unknown;
    integer: Submodule<stringInteger.submodule>;
    date: Submodule<stringDate.submodule>;
    digits: unknown;
    email: unknown;
    ip: Submodule<ip.submodule>;
    numeric: Submodule<stringNumeric.submodule>;
    semver: unknown;
    uuid: Submodule<{
        root: string.matching<string> | "00000000-0000-0000-0000-000000000000" | "ffffffff-ffff-ffff-ffff-ffffffffffff";
        v4: unknown;
        v6: unknown;
        v1: unknown;
        v2: unknown;
        v3: unknown;
        v5: unknown;
        v7: unknown;
        v8: unknown;
    }>;
    upper: Submodule<upper.submodule>;
    lower: Submodule<lower.submodule>;
    capitalize: Submodule<capitalize.submodule>;
}>;
export type Matching<rule> = {
    matching: constraint<rule>;
};
export declare namespace string {
    type atLeastLength<rule> = of<string, AtLeastLength<rule>>;
    type moreThanLength<rule> = of<string, MoreThanLength<rule>>;
    type atMostLength<rule> = of<string, AtMostLength<rule>>;
    type lessThanLength<rule> = of<string, LessThanLength<rule>>;
    type exactlyLength<rule> = of<string, ExactlyLength<rule>>;
    type matching<rule> = of<string, Matching<rule>>;
    type narrowed = of<string, Narrowed>;
    type optional = of<string, Optional>;
    type defaultsTo<rule> = of<string, Default<rule>>;
    type branded<rule> = of<string, Predicate<rule>>;
    type is<attributes> = of<string, attributes>;
    type applyAttribute<attribute> = attribute extends ExactlyLength<infer rule> ? exactlyLength<rule> : attribute extends MoreThanLength<infer rule> ? moreThanLength<rule> : attribute extends AtLeastLength<infer rule> ? atLeastLength<rule> : attribute extends AtMostLength<infer rule> ? atMostLength<rule> : attribute extends LessThanLength<infer rule> ? lessThanLength<rule> : attribute extends Matching<infer rule> ? matching<rule> : attribute extends Optional ? optional : attribute extends Default<infer rule> ? defaultsTo<rule> : attribute extends Predicate<infer rule> ? branded<rule> : never;
    type module = Module<string.submodule>;
    type submodule = Submodule<$>;
    type $ = {
        root: string;
        alpha: alpha;
        alphanumeric: alphanumeric;
        base64: base64.submodule;
        digits: digits;
        numeric: stringNumeric.submodule;
        integer: stringInteger.submodule;
        creditCard: creditCard;
        email: email;
        uuid: uuid.submodule;
        semver: semver;
        ip: ip.submodule;
        json: stringJson.submodule;
        date: stringDate.submodule;
        url: url.submodule;
        trim: trim.submodule;
        normalize: normalize.submodule;
        capitalize: capitalize.submodule;
        lower: lower.submodule;
        upper: upper.submodule;
    };
}
export type { brandedString } from "./brands.ts";
